import numpy as np
from billions.tools.log import *

class my_profolio:

    def __init__(self, date, total_amount=1000000, fluent=1) -> None:
        self.profolio = {'freemoney': total_amount}
        self.fluent = fluent
        self.date = date

    def get_profolio(self):
        pro = list(self.profolio.keys())
        pro.remove('freemoney')
        return pro

    def buy(self, stock, amount):
        if self.profolio['freemoney'] < amount:
            log(str(self.date) + ' -- Sorry, you don\'t have enoung money')
        else:
            if stock in self.profolio.keys():
                self.profolio[stock] += amount
                self.profolio['freemoney'] -= amount
            else:
                self.profolio[stock] = amount
                self.profolio['freemoney'] -= amount

    def sell(self, stock, amount):
        if stock not in self.profolio.keys():
            log(str(self.date) + ' -- Sorry, you don\'t have ' + stock)
        else:
            if self.profolio[stock] < amount:
                self.sell_all([stock])
            else:
                self.profolio[stock] -= amount
                self.profolio['freemoney'] += amount
                if self.profolio[stock] == 0:
                    self.profolio.pop(stock)

    def profolio_report(self):
        log(str(self.date) + ' -- my profolio: ' + str(self.profolio) +
              ' -- total: ' + str(np.mean(list(self.profolio.values()))))

    def adjust_price(self, date, rets):
        self.date = date
        for code in self.profolio.keys():
            if code != 'freemoney':
                self.profolio[code] = self.profolio[code] * ((rets[code])+1)

    def sell_all(self, stocks):
        for stock in stocks:
            if stock in self.profolio.keys():
                self.sell(stock, self.profolio[stock])
            else:
                log(str(self.date) + ' -- Sorry, you don\'t have this ' + stock)


class euqal_profolio(my_profolio):

    def __init__(self, date, total_amount, fluent) -> None:
        super().__init__(date, total_amount, fluent)

    def add(self, stocks):
        for stock in stocks:
            self.profolio[stock] = 0

    def add_n_sell(self, buys, sells):
        self.sell_all(sells)
        self.add(buys)
        self.rebalance()

    def rebalance(self):
        each = int(self.fluent*np.sum(list(self.profolio.values())) /
                   (len(self.profolio)-1))
        for s in self.profolio.keys():
            if s == 'freemoney':
                pass
            else:
                if self.profolio[s] > each:
                    self.sell(s, self.profolio[s] - each)
        for s in self.profolio.keys():
            if s == 'freemoney':
                pass
            else:
                if self.profolio[s] < each:
                    self.buy(s, each - self.profolio[s])



import pandas as pd
import numpy as np
from pandas.io.sql import read_sql_query
from tqdm import tqdm

from billions.tools.profolio import *
from billions.tools.deal_time import *
from billions.tools.ploter import *

# TODO: 手续费
# TODO：买入和卖出价格，当前均使用收盘价，假设能够以收盘价成交s


class trader:
    """
    This is the parent class for all trader class. In fact, you can use it to test many strategy.

    args:
    data: data you need
    amount: money you have in your account.
    profolio: profolio type (myprofolio, euqal_profolio)
    peride: the peride you need to change your profolio
    """

    def __init__(self, data, total_amount=1000000, fluent=1, profolio=euqal_profolio, peride='D', name='backtest'):
        """
        init function
        """
        self.type = profolio
        self.total_amount = total_amount
        self.fluent = fluent
        self.peride = peride
        self.name = name
        self.lables = [1]
        self.dates = np.unique(np.array(list(data.index))[:, 0])
        self.profolio = profolio(self.dates[0], total_amount, fluent)

        # self.ret = self.caculate_return()

        self.results = {}

    # mid function

    def get_trade_date(self):
        """
        **You don't want to change this**
        it can return the trade date you need.
        """
        if type(self.peride) == int:
            self.trade_dates = ndays(self.dates, self.peride)
        else:
            if self.peride == 'D':
                self.trade_dates = self.dates
            if self.peride == 'M':
                self.trade_dates = monthly(self.dates)
            if self.peride == 'Mf':
                self.trade_dates = monthly(self.dates, type='first')
            if self.peride == 'Ml':
                self.trade_dates = monthly(self.dates, type='last')

    def today_return(self, date):
        """
        it can get returns in certain date
        **You don't want to change this**
        """
        r = self.ret.loc[date].reset_index()[['code', 'return']]
        return dict(r.set_index('code')['return'])

    # set function
    def set_lable(self, lable):
        """
        change the lable
        **You don't want to change this**
        """
        if lable != -1:
            self.lable = lable

    def set_peride(self, peride):
        """
        change the peride
        **You don't want to change this**
        """
        if peride != -1:
            self.peride = peride

    def reset_profolio(self, reset):
        """
        reset the profolio
        **You don't want to change this**
        """
        if reset:
            self.profolio = self.type(
                self.dates[0], self.total_amount, self.fluent)

    # analysis function
    def ploter(self):
        """
        return the ploter
        **You don't want to change this**
        """
        return theploter(self)

    # main function
    def _run(self, lable=-1, peride=-1, reset=True):
        """
        it go over every day and get result
        **You don't want to change this**

        return:
        result: everyday profolio and amount
        """
        # set basic info
        self.set_lable(lable)
        self.set_peride(peride)
        self.reset_profolio(reset)

        # get trade date
        self.get_trade_date()

        # satrt to trade
        result = {}
        for date in tqdm(self.dates, leave=False, desc="lable-" + str(self.lable)):
            ret = self.today_return(date)
            self.profolio.adjust_price(date, ret)  # adjust daily price
            if date in self.trade_dates:
                self.trade(date)
            result[date] = self.profolio.profolio.copy()
        result = pd.DataFrame(result.values(), index=result.keys())
        result = result.fillna(0)
        self.results[self.lable] = result

        result = self.results[self.lable]
        self.networth = pd.DataFrame(result.sum(axis=1))
        self.networth.columns = ['networth']
        return self.networth

    def run(self, peride=-1):
        """
        run for all lable
        **You don't want to change this**
        """
        networths = []
        for lable in self.lables:
            networth = self._run(lable=lable, peride=peride, reset=True)
            networths.append(networth)

        # get networth
        self.networth = pd.concat(networths, axis=1)
        self.networth.columns = list(
            map(lambda x: 'group-' + str(x), self.lables))

        # get long short
        if len(self.lables) > 4:
            self.longshort = self.networth.copy()
            self.longshort['long/short-1'] = self.networth['group-' +
                                                           str(1)] / self.networth['group-'+str(self.number)]
            self.longshort['long/short-2'] = self.networth['group-' +
                                                           str(2)] / self.networth['group-'+str(self.number-1)]
            for i in range(1, self.number+1, 1):
                self.longshort = self.longshort.drop('group-'+str(i), axis=1)
        elif len(self.lables) > 2:
            self.longshort = self.networth.copy()
            self.longshort['long/short-1'] = self.networth['group-' +
                                                           str(1)] / self.networth['group-'+str(self.number)]
            for i in range(1, self.number+1, 1):
                self.longshort = self.longshort.drop('group-'+str(i), axis=1)
        elif len(self.lables) < 2:
            self.longshort = -1

    def trade(self, date):
        pass

    def signal(self):
        return

    def caculate_return(self):
        return


class group_trader(trader):
    """
    This is the parent class for all trader class. In fact, you can use it to test many strategy.

    !!! : caculator return all using close price.

    parent args:
    data: data you need
    amount: money you have in your account.
    profolio: profolio type (myprofolio, euqal_profolio)
    peride: the peride you need to change your profolio

    unique args:
    number: the number of groups
    inverse: if the factor is inversed
    """

    def __init__(self, data, total_amount=1000000, fluent=1, profolio=euqal_profolio, peride='D', number=5, inverse=False, name='backtest'):
        """
        init fuction
        """
        trader.__init__(self, data, total_amount,
                        fluent, profolio, peride, name)

        self.factor = data[['factor']]
        self.close = data[['close']]
        self.inverse = not inverse
        self.number = number

        self.lables = list(range(1, self.number+1, 1))
        self.ret = self.caculate_return()

    # main function
    def trade(self, date):
        """
        trade function: change profolio to target
        """
        codes = self.signal(date)
        buys = []
        sells = []
        for code in self.profolio.get_profolio():
            if code not in codes:
                sells.append(code)
        for code in codes:
            if code not in self.profolio.get_profolio():
                buys.append(code)

        self.profolio.add_n_sell(buys, sells)

    # child tool function
    def signal(self, date):
        """
        get code list in a certain date
        """
        today_factor = self.factor.loc[date, :].sort_values(
            'factor', ascending=self.inverse).reset_index()
        if self.inverse:
            def change(x): return -x
        else:
            def change(x): return x
        today_factor['factor'] = pd.qcut(
            change(today_factor['factor']), self.number, labels=range(1, self.number+1, 1))
        codes = list(
            today_factor[today_factor['factor'] == self.lable]['code'])
        return codes

    def caculate_return(self):
        """
        get return from close
        """
        c = self.close.reset_index().pivot('trade_dt', 'code')['close']
        ret = (c - c.shift())/c.shift()
        ret = ret.fillna(0)
        ret = pd.DataFrame(ret.stack())
        ret.columns = ['return']
        return ret



import pandas as pd


def __init__():
    pass


def loader_factor_close(fname, cname, ftype='csv'):
    """
    Read data from file.

    args:
        fname: factor file name (with columns: [factor, trade_dt, code])
        cname: close file name (with columns: [close, trade_dt, code])
        type: file type ([csv, xlsx, json])

    returns:
        data: 2-index(trade_dt, code), factor, close
    """

    # check out filetype
    if ftype == 'csv':
        read = pd.read_csv
    elif ftype == 'xlsx':
        read = pd.read_excel
    elif ftype == 'json':
        read = pd.read_json
    else:
        print('TYPE NOT SUPPORTED! Try other types like:')
        print(['csv', 'xlsx', 'json'])
        return

    # read data
    try:
        factor = read(fname)
        close = read(cname)
    except ValueError:
        print('WRONG TYPE, please check out your file type')
        return

    # choose the right columns
    factor = factor[['trade_dt', 'code', 'factor']]
    close = close[['trade_dt', 'code', 'close']]

    # combine two data set into one
    data = factor.merge(close, how='inner')
    data.trade_dt = pd.DatetimeIndex(data.trade_dt)
    data = data.set_index(['trade_dt', 'code'])

    return data


def load_price(fname):
    data = pd.read_csv(fname, index_col=0)
    data = data.set_index(['trade_dt', 'code'])
    return data



import pandas as pd
import plotly
import plotly.express as px
import pandas as pd
from plotly.subplots import make_subplots
import plotly.graph_objects as go


class theploter:
    def __init__(self, trader):
        self.trader = trader

    def plot_networth(self):
        fig = px.line(self.trader.networth)
        fig.update_layout(title=self.get_name())
        fig.show()

    def plot_long_short(self):
        fig = px.line(self.trader.longshort)
        fig.update_layout(title=self.get_name())
        fig.show()

    def plot(self):
        if type(self.trader.longshort) == int:
            self.plot_networth()
        else:
            fig = plotly.subplots.make_subplots(rows=1, cols=2)

            for column in self.trader.networth.columns:
                fig.append_trace(go.Scatter(
                    x=self.trader.networth.index, y=self.trader.networth[column], name=column), 1, 1)
            for column in self.trader.longshort.columns:
                fig.append_trace(go.Scatter(
                    x=self.trader.longshort.index, y=self.trader.longshort[column], name=column), 1, 2)
            fig.update_layout(title=self.get_name())
            fig.show()

    def get_name(self):
        if type(self.trader.peride) == str:
            if self.trader.peride == 'D':
                return self.trader.name + ' -- ' + 'daily backtest'
            if self.trader.peride[0] == 'M':
                return self.trader.name + ' -- ' + 'monthly backtest'
        if type(self.trader.peride) == int:
            return self.trader.name + ' -- ' + str(self.trader.peride) + ' days backtest'



import pandas as pd


def monthly(data, type='first'):
    data = pd.DataFrame(data)
    data.columns = ['date']
    data['year'] = data['date'].apply(lambda x: x.year)
    data['month'] = data['date'].apply(lambda x: x.month)
    grouped = data['date'].groupby([data['year'], data['month']])
    if type == 'first':
        last_day = grouped.min().to_list()
    elif type == 'last':
        last_day = grouped.max().to_list()
    return last_day


def ndays(data, n):
    data = pd.DataFrame(data)
    thedates = list(data.loc[range(0, len(data), n)][0])
    return thedates



#!/user/bin/env python
# coding: utf-8

from setuptools import _install_setup_requires, setup, find_packages

setup(
    name='Shouzheng_billions',
    version = '0.0.2',
    author='Shouzheng-Hu',
    author_email='obobob32@gmail.com',
    url='https://github.com/wintemmer/billions',
    packages=find_packages()
)



def log(txt):
    file = r'log.txt'
    with open(file, 'a+') as f:
        f.write(txt+'\n')  # 加\n换行显示

def start():
    file = r'log.txt'
    with open(file, 'a+') as f:
        f.write('The backtest begins'+'\n')  # 加\n换行显示
